#include "CompositeActor.h"

CompositeActor::CompositeActor(
	Console* console,
	FallingDownActor* fallingDownActor,
	SingleActionActor* hitRecoveringActor,
	MovingActor* movingActor,
	SingleActionActor* singleHitActor,
	ComboActor* comboActor,
	SingleActionActor* healActor,
	SingleActionActor* mysticalActor,
	ChargingActor* chargingActor
	) {
	
	m_console = console;
	m_fallingDownActor = fallingDownActor;
	m_hitRecoveringActor = hitRecoveringActor;
	m_movingActor = movingActor;
	m_singleHitActor = singleHitActor;
	m_comboActor = comboActor;
	m_healActor = healActor;
	m_mysticalActor = mysticalActor;
	m_chargingActor = chargingActor;
	
	m_hasDied = false;
	m_canBeRemoved = false;
	
	m_console->Debug( "CompositeActor: at %p", this );
}

CompositeActor::~CompositeActor() {
	delete m_chargingActor;
	delete m_mysticalActor;
	delete m_healActor;
	delete m_comboActor;
	delete m_singleHitActor;
	delete m_fallingDownActor;
	delete m_hitRecoveringActor;
	delete m_movingActor;
	m_console->Debug( "~CompositeActor: at %p", this );
}

bool CompositeActor::IsInterruptibleBySelf() {
	bool fallingDownActor = m_fallingDownActor->IsInterruptible();
	bool movingActor = m_movingActor->IsInterruptible();
	bool hitRecoveringActor = m_hitRecoveringActor->IsInterruptibleBySelf();
	bool singleHitActor = m_singleHitActor->IsInterruptibleBySelf();
	bool comboActor = m_comboActor->IsInterruptibleBySelf();
	
	bool healActor = m_healActor->IsInterruptibleBySelf();
	bool mysticalActor = m_mysticalActor->IsInterruptibleBySelf();
	
	bool chargingActor = m_chargingActor->IsInterruptibleBySelf();
	
	bool result = !m_canBeRemoved &&
					chargingActor &&
					mysticalActor &&
					healActor &&
					comboActor &&
					singleHitActor &&
					fallingDownActor &&
					hitRecoveringActor &&
					movingActor;
	
	return result;
}

bool CompositeActor::IsInterruptibleByOther() {
	bool fallingDownActor = m_fallingDownActor->IsInterruptible();
	bool movingActor = m_movingActor->IsInterruptible();
	bool hitRecoveringActor = m_hitRecoveringActor->IsInterruptibleByOther();
	bool singleHitActor = m_singleHitActor->IsInterruptibleByOther();
	bool comboActor = m_comboActor->IsInterruptibleByOther();
	
	bool healActor = m_healActor->IsInterruptibleByOther();
	bool mysticalActor = m_mysticalActor->IsInterruptibleByOther();
	
	bool chargingActor = m_chargingActor->IsInterruptibleByOther();
	
	bool result = !m_canBeRemoved &&
					chargingActor &&
					mysticalActor &&
					healActor &&
					comboActor &&
					singleHitActor &&
					fallingDownActor &&
					hitRecoveringActor &&
					movingActor;
	
	return result;
}

bool CompositeActor::CanMoveOrStand() {
	return this->IsInterruptibleBySelf();
}

void CompositeActor::MoveBy( Coords_F deltaPosition ) {
	if( !this->CanMoveOrStand() ) {
		return;
	}
	m_movingActor->MoveBy( deltaPosition );
}

void CompositeActor::Stand() {
	if( !this->CanMoveOrStand() ) {
		return;
	}
	
	m_movingActor->Stand( false );
}



bool CompositeActor::CanFallDownOrRecoverFromHit() {
	return this->IsInterruptibleByOther();
}

void CompositeActor::FallDown() {
	if( !this->CanFallDownOrRecoverFromHit() ) {
		return;
	}
	this->CancelAllActors();
	m_fallingDownActor->FallDown( m_displacementTendencyIsToTheRight );
}

void CompositeActor::RecoverFromHit() {
	if( !this->CanFallDownOrRecoverFromHit() ) {
		return;
	}
	this->CancelAllActors();
	m_hitRecoveringActor->PerformAction();
}


bool CompositeActor::CanPerformSingleHit() {
	return this->IsInterruptibleBySelf();
}

void CompositeActor::PerformSingleHit(
	Event* animationProgressReachedEvent,
	float animationProgress
	) {
	if( !this->CanPerformSingleHit() ) {
		return;
	}
	this->CancelAllActors();
	m_singleHitActor->PerformAction(
		animationProgressReachedEvent,
		animationProgress
		);
}


bool CompositeActor::CanPerformCombo() {
	return this->IsInterruptibleBySelf();
}

void CompositeActor::PerformCombo(
	Event* continuationEvent,
	Event* animationProgressReachedEvent,
	float animationProgress
	) {
	if( !this->CanPerformCombo() ) {
		return;
	}
	this->CancelAllActors();
	m_comboActor->PerformCombo(
		continuationEvent,
		animationProgressReachedEvent,
		animationProgress
		);
}


bool CompositeActor::CanHealOrMystical() {
	bool isBeingHit = !m_hitRecoveringActor->IsInterruptibleBySelf() || 
					!m_hitRecoveringActor->IsInterruptibleByOther();
	// allow heal or mystical when recovering from a hit, in order
	// to make it more lenient to use in tight situations
	return this->IsInterruptibleBySelf() || isBeingHit;
}
void CompositeActor::Heal() {
	if( !this->CanHealOrMystical() ) {
		return;
	}
	this->CancelAllActors();
	m_healActor->PerformAction();
}
void CompositeActor::Mystical() {
	if( !this->CanHealOrMystical() ) {
		return;
	}
	this->CancelAllActors();
	m_mysticalActor->PerformAction();
}
bool CompositeActor::IsPerformingMystical() {
	return !m_mysticalActor->IsInterruptibleBySelf() ||
		   !m_mysticalActor->IsInterruptibleByOther();
}

bool CompositeActor::CanCharge() {
	return this->IsInterruptibleBySelf();
}

void CompositeActor::Charge( 
	Coords_F destination,
	float lengthUnitsPerFrame
	) {
	if( !this->CanCharge() ) {
		return;
	}
	this->CancelAllActors();
	m_chargingActor->Charge( destination, lengthUnitsPerFrame );
}

bool CompositeActor::IsCharging() {
	return !m_chargingActor->IsInterruptibleBySelf() ||
		   !m_chargingActor->IsInterruptibleByOther();
}


bool CompositeActor::TryRenderFallingDownActor() {
	// not active?
	if( m_fallingDownActor->IsInterruptible() ) {
		return false;
	}
	
	m_fallingDownActor->Render();
	
	// transition out of a falling down state
	if( m_fallingDownActor->IsInterruptible() ) {
		if( m_hasDied ) {
			m_canBeRemoved = true;
			// move offscreen
			Coords_F offscreen( 256.0f, 0.0f );
			m_movingActor->RelocateTo( offscreen );
			m_movingActor->Render();
		} else {
			m_movingActor->GetUp();
		}
	}
	
	return true;
}

bool CompositeActor::TryRenderHitRecoveringActor() {
	return Helper_TryRenderSingle( m_hitRecoveringActor );
}

bool CompositeActor::TryRenderSingleHitActor() {
	return Helper_TryRenderSingle( m_singleHitActor );
}

bool CompositeActor::TryRenderComboActor() {
	
	bool isInterruptible = 
		m_comboActor->IsInterruptibleBySelf() &&
		m_comboActor->IsInterruptibleByOther();

	// not active?
	if( isInterruptible ) {
		return false;
	}
	
	m_comboActor->Render();
	
	// check again
	isInterruptible = 
		m_comboActor->IsInterruptibleBySelf() &&
		m_comboActor->IsInterruptibleByOther();
	
	// have we just come out of the combo state?
	if( isInterruptible ) {
		m_movingActor->Stand( true );
	}
	
	return true;
}

bool CompositeActor::TryRenderHealActor() {
	return Helper_TryRenderSingle( m_healActor );
}

bool CompositeActor::TryRenderMysticalActor() {
	return Helper_TryRenderSingle( m_mysticalActor );
}

bool CompositeActor::Helper_TryRenderSingle( SingleActionActor* actor ) {
	bool isInterruptible = 
		actor->IsInterruptibleBySelf() &&
		actor->IsInterruptibleByOther();

	// not active?
	if( isInterruptible ) {
		return false;
	}
	
	actor->Render();
	
	// check again
	isInterruptible = 
		actor->IsInterruptibleBySelf() &&
		actor->IsInterruptibleByOther();
	
	// transition out of the action state
	if( isInterruptible ) {
		m_movingActor->Stand( true );
	}
	
	return true;
}

bool CompositeActor::TryRenderChargingActor() {
	bool isInterruptible = 
		m_chargingActor->IsInterruptibleBySelf() &&
		m_chargingActor->IsInterruptibleByOther();

	// not active?
	if( isInterruptible ) {
		return false;
	}
	
	m_chargingActor->Render();
	
	// check again
	isInterruptible = 
		m_chargingActor->IsInterruptibleBySelf() &&
		m_chargingActor->IsInterruptibleByOther();
	
	// transition out of the action state
	if( isInterruptible ) {
		m_movingActor->Stand( true );
	}
	
	return true;
}

void CompositeActor::Render() {
	if( m_canBeRemoved ) {
		return;
	}
	
	if( TryRenderFallingDownActor() ) {
		return;
	}
	
	if( TryRenderHitRecoveringActor() ) {
		return;
	}
	
	if( TryRenderSingleHitActor() ) {
		return;
	}
	
	if( TryRenderComboActor() ) {
		return;
	}
	
	if( TryRenderHealActor() ) {
		return;
	}
	
	if( TryRenderMysticalActor() ) {
		return;
	}
	
	if( TryRenderChargingActor() ) {
		return;
	}
	
	m_movingActor->Render();
}

void CompositeActor::CancelAllActors() {
	m_fallingDownActor->Cancel();
	m_hitRecoveringActor->Cancel();
	m_movingActor->Cancel();
	m_singleHitActor->Cancel();
	m_comboActor->Cancel();
	m_chargingActor->Cancel();
}

void CompositeActor::RelocateTo( Coords_F position ) {
	m_movingActor->RelocateTo( position );
}

void CompositeActor::SetIsFacingRight( bool isFacingRight ) {
	m_movingActor->SetIsFacingRight( isFacingRight );
}

bool CompositeActor::CanBeRemoved() {
	return m_canBeRemoved;
}

void CompositeActor::SetDisplacementTendency( bool isTowardsRight ) {
	m_displacementTendencyIsToTheRight = isTowardsRight;
}

void CompositeActor::SetHasDied() {
	m_hasDied = true;
}