#include "MovingActor.h"

MovingActor::MovingActor( 
	Console* console,
	Animator* animator,
	bool isFacingRight
	) {
	
	m_console = console;
	m_animator = animator;
	
	m_isStanding = false;
	m_isGettingUp = false;
	
	m_getUpAnimationOverEvent = NULL;
	
	m_animator->GetSprite()->SetIsFacingRight( isFacingRight );
	
	// a good initial state
	this->Stand( true );
}

MovingActor::~MovingActor() {
	delete m_animator;
	
	if( NULL != m_getUpAnimationOverEvent ) {
		delete m_getUpAnimationOverEvent;
	}
}

void MovingActor::MoveBy( Coords_F deltaPosition ) {

	// can't move if we're getting up
	if( m_isGettingUp ) {
		return;
	}
	
	// are we coming out of a standing state?
	if( m_isStanding ) {
		m_animator->ChangePhase( PHASE_WALKING );
	}

	OAMSprite* sprite = m_animator->GetSprite();
	
	Coords_F newPosition = sprite->GetPosition() + deltaPosition;
	sprite->MoveTo( newPosition );
		
	if( deltaPosition.x < 0 ) {
		sprite->SetIsFacingRight( false );
	} else if( deltaPosition.x > 0 ) {
		sprite->SetIsFacingRight( true );
	} // if it hasn't changed horizontal position, keep previous orientation
	
	m_isStanding = false;
}

void MovingActor::Stand( bool resetAnimation ) {
	
	// are we already standing, and we don't wish to reset the animation?
	if( m_isStanding && !resetAnimation ) {
		return;
	}
	
	// can start standing up if getting up
	if( m_isGettingUp ) {
		return;
	}
	
	m_animator->ChangePhase( PHASE_STANDING );
	m_isStanding = true;
}

void MovingActor::GetUp() {
	
	if( m_isGettingUp ) {
		return;
	}
	
	if( NULL != m_getUpAnimationOverEvent ) {
		delete m_getUpAnimationOverEvent;
	}
	m_getUpAnimationOverEvent = new Event();
	
	m_animator->ChangePhase( 
		PHASE_GETTING_UP,
		m_getUpAnimationOverEvent
		);
	
	m_isGettingUp = true;
	m_isStanding = false;
}

void MovingActor::Render() {
	m_animator->AdvanceFrame();

	// getting up is uninterruptible, so finish it if it's in progress
	if( m_isGettingUp && m_getUpAnimationOverEvent->IsSet() ) {
		// we've finished getting up
		this->Cancel();
	}
	
	m_animator->Render();
}

bool MovingActor::IsInterruptible() {
	return !m_isGettingUp;
}

void MovingActor::Cancel() {
	if( NULL != m_getUpAnimationOverEvent ) {
		delete m_getUpAnimationOverEvent;
		m_getUpAnimationOverEvent = NULL;
	}
	
	m_isGettingUp = false;	
	this->Stand( true );
}

Coords_F MovingActor::GetPosition() {
	return m_animator->GetSprite()->GetPosition();
}

bool MovingActor::IsFacingRight() {
	return m_animator->GetSprite()->GetIsFacingRight();
}

void MovingActor::SetIsFacingRight( bool isFacingRight ) {
	m_animator->GetSprite()->SetIsFacingRight( isFacingRight );
}

Coords_F MovingActor::GetCentrePosition() {
	return m_animator->GetSprite()->GetCentrePosition();
}

void MovingActor::RelocateTo( Coords_F position ) {
	m_animator->GetSprite()->MoveToUnchecked( position );
}