#include "OAMSprite.h"

OAMSprite::OAMSprite( 
	u8 OAMSpriteIndex, 
	u8* graphicsData, 
	u16 spriteSheetWidthInPixels,
	SpritePool* spritePool,
	Console* console,
	OamState* oam
	) {
	
	m_myOAM = oam;
	m_console = console;
	m_spritePool = spritePool;
	m_priority = OBJPRIORITY_2;
	
	this->SetOAMSpriteIndex( OAMSpriteIndex, false );

	// get a slot in the OAM
	m_spriteCurrentFrameGraphicsData = oamAllocateGfx(
		m_myOAM, 
		m_spriteSizeEnum, 
		SpriteColorFormat_256Color
		);
	
	if( NULL == m_spriteCurrentFrameGraphicsData ) {
		m_console->Error( "OAMSprite: Memory bank full!" );
	}
	
	m_spriteAllGraphicsData = (u8*)graphicsData;
	m_spriteSheetWidthInFrames = spriteSheetWidthInPixels / SPRITE_FRAME_SIDE_LENGTH;
	
	// these values will ensure reprioritization 
	// during the first frame this sprite is on-screen
	m_oldPosition.x = 400;
	m_oldPosition.y = 400;
	m_position.x = 300;
	m_position.y = 300;
	
	this->ChangeFrame( 0, 0 );
	
	m_isFacingRight = true;
	
	m_console->Debug( "OAMSprite: %p", this );
}

OAMSprite::~OAMSprite() {

	this->ForceOffScreen();

	// free up OAM taken up by this sprite's current frame
	oamFreeGfx( 
		m_myOAM, 
		m_spriteCurrentFrameGraphicsData
		);
		
	// tell the sprite manager to remove its records as well
	m_spritePool->FreeSlot( this );
	
	m_console->Debug( "~OAMSprite: %p", this );
}

void OAMSprite::ChangeFrame( u8 row, u8 column ) {

	// transform row and column of desired 64x64 square into a single co-ordinate like so:
	// (assume a 4 row, 3 column sheet, and desired square is on row 1, column 2)
	// [ ][ ][ ]
	// [ ][ ][x]
	// [ ][ ][ ]
	// [ ][ ][ ]
	// 1-dimensional square co-ordinate = column + row*width = 2 + 1*3 = 5th from beginning
	//
	// it's worth mentioning that the bytes for a tile are contiguous in memory; this is due
	// to GRIT being told that the height and widht of a "metatile" is H and W, expressed in
	// "GRIT tile" heights and widths, which are 8 pixels each by default.
	// this allows for tiles to be loaded more easily
	u16 squareCoordinate = column + row * m_spriteSheetWidthInFrames;
	
	u8* offsetFromSpriteGraphicsDataOrigin = 
		m_spriteAllGraphicsData + squareCoordinate * SPRITE_FRAME_SIDE_LENGTH_SQUARED;
	
	dmaCopy( 
		offsetFromSpriteGraphicsDataOrigin, 
		m_spriteCurrentFrameGraphicsData, 
		SPRITE_FRAME_SIDE_LENGTH_SQUARED 
		);
}

void OAMSprite::Render() {

	this->WriteToOAM(
		m_OAMSpriteIndex,
		(s16)m_position.x, 
		(s16)m_position.y, 
		m_OAMExtendedPaletteIndex
		);
	
	m_oldPosition = m_position;
}

void OAMSprite::SetOAMSpriteIndex( 
	u8 oamSpriteIndex,
	bool forceOffScreen
	) {
	
	if( forceOffScreen ) {
		this->ForceOffScreen();
	}
	
	m_OAMExtendedPaletteIndex = oamSpriteIndex;
	m_OAMSpriteIndex = oamSpriteIndex;
}

void OAMSprite::MoveTo( Coords_F position ) {
	position.x = fmaxf( position.x, LEFT_BOUND );
	position.x = fminf( position.x, RIGHT_BOUND );
	position.y = fmaxf( position.y, TOP_BOUND );
	position.y = fminf( position.y, BOTTOM_BOUND );
	m_position = position;
}

void OAMSprite::MoveToUnchecked( Coords_F position ) {
	m_position = position;
}

Coords_F OAMSprite::GetPosition() {
	return m_position;
}

Coords_F OAMSprite::GetCentrePosition() {
	Coords_F centre = this->GetPosition();
	centre.x = centre.x + 31;
	centre.y = centre.y + 31;
	return centre;
}

Coords_F OAMSprite::GetPositionDelta() {
	return m_position - m_oldPosition;
}

void OAMSprite::SetIsFacingRight( bool isFacingRight ) {
	m_isFacingRight = isFacingRight;
}

bool OAMSprite::GetIsFacingRight() {
	return m_isFacingRight;
}

void OAMSprite::ForceOffScreen() {
	this->WriteToOAM(
		m_OAMSpriteIndex,
		256, 
		0, 
		m_OAMExtendedPaletteIndex
		);
}

void OAMSprite::SetPriorityHigh() {
	m_priority = OBJPRIORITY_1;
}

void OAMSprite::WriteToOAM(
	u8 index,
	s16 x,
	s16 y,
	u8 paletteIndex
	) {
	
	oamSet(
		m_myOAM, 
		index,
		x, 
		y, 
		m_priority, 
		paletteIndex,
		m_spriteSizeEnum, 
		SpriteColorFormat_256Color, 
		m_spriteCurrentFrameGraphicsData, 
		-1, 
		false, 
		false, 
		!m_isFacingRight, 
		false, 
		false
		);
}