#include "PrioritizingSpritePool.h"
#include "DataTypes.h"

PrioritizingSpritePool::PrioritizingSpritePool(
	Console* console,
	u8 capacity
	) : SpritePool( console, capacity ) 
	{
	
	m_console->Debug( "PrioritizingSpritePool: %p", this );
}

PrioritizingSpritePool::~PrioritizingSpritePool() {
	m_console->Debug( "~PrioritizingSpritePool: %p", this );
}

void PrioritizingSpritePool::FreeSlot( OAMSprite* sprite ) {
	
	SpritePool::FreeSlot( sprite );
	
	m_console->Debug( "[PrioritizingSpritePool]FreeSlot: %p", this );
	
	// get rid of nulls by compacting array
	// the point of this is to make reprioritizations faster, as 
	// Reprioritize happens much more often than FreeSlot
	
	OAMSprite** tempArray = new OAMSprite* [ m_capacity ];
	u8 tempIndex = 0;
	// copy non-NULL elements into temp array
	for( u8 i = 0; i < m_capacity; i++ ) {
		if( NULL != m_sprites[ i ] ) {
			tempArray[ tempIndex ] = m_sprites[ i ];
			tempIndex++;
		}
	}

	// NULL out remainder of temp array
	while( tempIndex < m_capacity ) {
		tempArray[ tempIndex ] = NULL;
		tempIndex++;
	}
	
	// copy temp array to our sprites array
	for( u8 i = 0; i < m_capacity; i++ ) {
		m_sprites[ i ] = tempArray[ i ];
	}
	
	this->SetAllOAMIndexes();
	
	delete tempArray;
}

void PrioritizingSpritePool::SetAllOAMIndexes() {
	for( u8 i = 0; i < m_capacity; i++ ) {

		if( NULL == m_sprites[ i ] ) {
			continue;
		}
		
		m_sprites[ i ]->SetOAMSpriteIndex( i, true );
	}
}

// This assumes that there are no NULL gaps in our array
// Gaps can only be introduced via FreeSlot
void PrioritizingSpritePool::Reprioritize() {
	
	// index 0 (highest priority) must have the largest Y value
	// index capacity-1 (lowest priority) must have the smallest Y value

	u8 nonNullCount = 0;
	
	// has any sprite moved vertically?
	bool anyVerticalMovement = false;
	for( u8 i = 0; i < m_capacity; i++ ) {
		if( NULL == m_sprites[ i ] ) {
			continue;
		}
		
		nonNullCount++;
		Coords_F positionDelta = m_sprites[ i ]->GetPositionDelta();
		if( 0.0f != positionDelta.y ) {
			anyVerticalMovement = true;
		}
	}
	
	// no - no reprioritization necessary
	if( !anyVerticalMovement ) {
		return;
	}
	
	// we must reprioritize
		
	// insertion sort our array, by Y values, descending
	OAMSprite* t;
	u8 previousRunner;
	for ( u8 i = 1; i < nonNullCount; i++ ) {
		previousRunner = i;
		
		while ( previousRunner > 0 ) {
		
			Coords_F position_back = m_sprites[ previousRunner - 1 ]->GetPosition();
			Coords_F position_current = m_sprites[ previousRunner ]->GetPosition();
			
			if( position_current.y <= position_back.y ) {
				break;
			}
				
			t = m_sprites[ previousRunner ];
			m_sprites[ previousRunner ] = m_sprites[ previousRunner - 1 ];
			m_sprites[ previousRunner - 1 ] = t;

			previousRunner--;
		}
	
	}
	
	this->SetAllOAMIndexes();
}

