#include "PlayerDriver.h"

PlayerDriver::PlayerDriver(
	CoreUtils* coreUtils,
	Console* console,
	Character* character,
	InputManager* inputManager,
	List<Character>* enemies,
	Scroller* scroller,
	Impact* impact
	) : Driver( coreUtils, console, character, impact ) {
	
	m_scroller = scroller;
	m_inputManager = inputManager;
	m_enemies = enemies;
	m_singleHitOccurredEvent.Reset();
	m_comboHitOccurredEvent.Reset();
	m_comboHitCount = 0;
	
	m_walkingSpeedUnitsPerFrame = character->GetProperties()->GetWalkingSpeed();
	m_hittingRadius = 36.0f;
	
	m_lastComboButtonPressFrame = 0;
	m_lastComboHitOccurrenceFrame = 0;
	m_frameCounter = 0;
	m_comboContinuationEvent.Reset();
}

PlayerDriver::~PlayerDriver() {
}

s16 PlayerDriver::AdjustDamage( s16 damage ) {

	// a full combo deals 8xdamage

	if( m_comboHitCount == 1 || m_comboHitCount == 2 ) {
		// first and second hits deal little damage
		damage = damage / 2;
	} else if( m_comboHitCount == 3 ) {
		// third hit deals more damage
		damage = damage * 2;
	} else if( m_comboHitCount == 4 ) {
		// last hit deals a lot of damage
		damage = damage * 5;
	}
	
	return damage;
}

s16 PlayerDriver::AdjustEnergyGain( s16 energyGain ) {
	if( m_comboHitCount == 1 || m_comboHitCount == 2 ) {
		// first and second hits award no energy
		energyGain = 0;
	} else if( m_comboHitCount == 4 ) {
		// last hit awards a lot of energy
		energyGain = energyGain * 7;
	}
	
	return energyGain;
}

// register a combo hit on enemies
bool PlayerDriver::TryComboHitNearbyEnemies() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	bool wasAnyoneHit = false;
	
	Node<Character>* node = m_enemies->GetHead();
	while( NULL != node ) {
		Character* enemy = node->GetData();
		UtilitiesCompositeActor* enemyActor = enemy->GetActor();
		
		if( myActor->IsFacingWithinDistanceOf( enemyActor, m_hittingRadius ) &&
			enemyActor->CanFallDownOrRecoverFromHit()
			) {
		
			s16 damage = AdjustDamage( m_character->GetProperties()->GetComboHitDamage() );
			s16 energyGain = AdjustEnergyGain( ENERGY_GAIN_PER_COMBO_HIT );
			
			enemy->GetProperties()->DecreaseLifeBy( damage );
			m_character->GetProperties()->GainEnergy( energyGain );
			m_character->SetLastVictim( enemy );
			bool hasDied = enemy->GetProperties()->GetLife() <= 0;
			
			enemyActor->SetDisplacementTendency( myActor->IsToTheLeftOf( enemyActor ) );
			
			// knock down on the last combo hit
			if( m_comboHitCount == 4 || hasDied ) {
				if( enemyActor->CanFallDownOrRecoverFromHit() ) {
					
					if( hasDied ) {
						enemyActor->SetHasDied();
					}
					enemyActor->FallDown();
					if( TryAllocateSound() ) {
						enemy->GetSoundboard()->WasThrownDown();
					}
				}
			} else {
				if( enemyActor->CanFallDownOrRecoverFromHit() ) {
					if( hasDied ) {
						enemyActor->SetHasDied();
					}
					enemyActor->RecoverFromHit();
				}
			}
			if( TryAllocateSound() ) {
				m_character->GetSoundboard()->ComboHitLanded();
			}
			wasAnyoneHit = true;
		}
		
		node = node->GetNext();
	}
	
	if( wasAnyoneHit ) {
		m_impact->ShowForPosition( myActor->GetPosition(), myActor->IsFacingRight() );
	}
	
	return wasAnyoneHit;
}

void PlayerDriver::ApplySingleHitOnNearbyEnemies() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	bool wasAnyoneHit = false;
	Node<Character>* node = m_enemies->GetHead();
	while( NULL != node ) {
		Character* enemy = node->GetData();
		UtilitiesCompositeActor* enemyActor = enemy->GetActor();
		float singleHitRadius = m_hittingRadius * 1.1f;
		if( myActor->IsFacingWithinDistanceOf( enemyActor, singleHitRadius ) &&
			enemyActor->CanFallDownOrRecoverFromHit()
			) {
			enemy->GetProperties()->DecreaseLifeBy( 
				m_character->GetProperties()->GetSingleHitDamage()
				);
			m_character->SetLastVictim( enemy );
			bool hasDied = enemy->GetProperties()->GetLife() <= 0;
			
			enemyActor->SetDisplacementTendency( myActor->IsToTheLeftOf( enemyActor ) );
			if( hasDied ) {
				enemyActor->SetHasDied();
			}
			
			if( TryAllocateSound() ) {
				m_character->GetSoundboard()->SingleHitLanded();
			}
			enemyActor->FallDown();
			
			if( TryAllocateSound() ) {
				enemy->GetSoundboard()->WasThrownDown();
			}
			wasAnyoneHit = true;
		}
		
		node = node->GetNext();
	}
	
	if( wasAnyoneHit ) {
		m_impact->ShowForPosition( myActor->GetPosition(), myActor->IsFacingRight() );
	}
}

void PlayerDriver::HandleSingleHit() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	
	// catch occurrence of a single hit
	if( m_singleHitOccurredEvent.IsSet() ) {
		m_console->Debug("Single hit event from player!");
		m_singleHitOccurredEvent.Reset();		
		
		this->ApplySingleHitOnNearbyEnemies();
		
		this->TurnCharacterIfRequested();
	}
	
	// begin a single hit
	if( myActor->CanPerformSingleHit() ) {
		s16 cost = m_character->GetProperties()->GetPowerStrikeEnergyCost();
		if( m_inputManager->IsButtonTapped( BUTTON_B ) &&
			m_character->GetProperties()->GetEnergy() >= cost
			) {
			m_character->GetProperties()->DecreaseEnergyBy( cost );
			myActor->PerformSingleHit(
				&m_singleHitOccurredEvent,
				0.5f
				);
			m_character->GetSoundboard()->StartSingleHit();
		}
	}
}

void PlayerDriver::HandleComboHit() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	bool buttonIsTapped = m_inputManager->IsButtonTapped( BUTTON_Y );
	
	// record combo button hit
	if( buttonIsTapped ) {
		m_lastComboButtonPressFrame = m_frameCounter;
	}
	
	u32 framesBetweenLastHitAndLastButtonPress = 
		m_lastComboButtonPressFrame - m_lastComboHitOccurrenceFrame;
	
	bool shouldContinueCombo = 
		framesBetweenLastHitAndLastButtonPress < FRAME_COUNT_ALLOWED_WITH_NO_COMBO_CONTINUATION_PRESS;

	if( m_lastHitOccurrenceHitSomeone && 
		shouldContinueCombo ) {
		
		m_comboContinuationEvent.Set();
	} else {
		m_comboContinuationEvent.Reset();
	}
	
	// catch occurrence of a combo hit
	if( m_comboHitOccurredEvent.IsSet() ) {
		m_console->Debug("Combo hit event from player!");
		m_comboHitCount++;
		m_comboHitOccurredEvent.Reset();
		m_lastComboHitOccurrenceFrame = m_frameCounter;
		
		m_lastHitOccurrenceHitSomeone = this->TryComboHitNearbyEnemies();
		
		this->TurnCharacterIfRequested();
	}
	
	// begin a combo hit
	if( myActor->CanPerformCombo() ) {
		if( buttonIsTapped ) {
			m_comboHitCount = 0;
			m_lastHitOccurrenceHitSomeone = false;
			m_comboContinuationEvent.Reset();
			myActor->PerformCombo(
				&m_comboContinuationEvent,
				&m_comboHitOccurredEvent,
				0.5f
				);
			m_character->GetSoundboard()->StartComboHit();
		}
	}
}

// turn if there's player input
// this improves the dynamism of the player character during attacks
void PlayerDriver::TurnCharacterIfRequested() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	if( m_inputManager->IsButtonPressed( DPAD_LEFT ) ) {
		myActor->SetIsFacingRight( false );
	} else if( m_inputManager->IsButtonPressed( DPAD_RIGHT ) ) {
		myActor->SetIsFacingRight( true );
	}
}

void PlayerDriver::HandleHeal() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	if( !myActor->CanHealOrMystical() ) {
		return;
	}
	
	if( !m_inputManager->IsButtonTapped( BUTTON_X ) ) {
		return;
	}
	
	CharacterProperties* properties = m_character->GetProperties();
	if( properties->GetEnergy() < properties->GetHealAndMysticalEnergyCost() ) {
		return;
	}
	
	m_character->GetSoundboard()->Heal();
	properties->HealAndSpendEnergy();
	myActor->Heal();
}

void PlayerDriver::HandleMystical() {

	UtilitiesCompositeActor* myActor = m_character->GetActor();
	
	if( myActor->IsPerformingMystical() ) {
		Node<Character>* node = m_enemies->GetHead();
		while( NULL != node ) {
			Character* enemy = node->GetData();
			UtilitiesCompositeActor* enemyActor = enemy->GetActor();
			
			if( enemyActor->CanFallDownOrRecoverFromHit() ) {
				enemy->GetProperties()->DecreaseLifeBy( 
					m_character->GetProperties()->GetMysticalDamage()
					);
				m_character->SetLastVictim( enemy );
				bool hasDied = enemy->GetProperties()->GetLife() <= 0;
				
				enemyActor->SetDisplacementTendency( myActor->IsToTheLeftOf( enemyActor ) );
				if( hasDied ) {
					enemyActor->SetHasDied();
				}
				
				enemyActor->FallDown();
				if( TryAllocateSound() ) {
					enemy->GetSoundboard()->WasThrownDown();
				}
			}
			
			node = node->GetNext();
		}
		return;
	}

	// should we start performing a mystical attack?
	if( !myActor->CanHealOrMystical() ) {
		return;
	}
	if( !m_inputManager->IsButtonTapped( BUTTON_A ) ) {
		return;
	}
	CharacterProperties* properties = m_character->GetProperties();
	if( properties->GetEnergy() < properties->GetHealAndMysticalEnergyCost() ) {
		return;
	}
	
	m_character->GetSoundboard()->Mystical();
	properties->DecreaseEnergyBy( 
		properties->GetHealAndMysticalEnergyCost()
		);
	myActor->Mystical();
}

void PlayerDriver::HandleMovement() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	if( !myActor->CanMoveOrStand() ) {
		return;
	}

	Coords_F down( 0.0f, m_walkingSpeedUnitsPerFrame );
	Coords_F up( 0.0f, -1.0f * m_walkingSpeedUnitsPerFrame );
	Coords_F left( -1.0f * m_walkingSpeedUnitsPerFrame, 0.0f );
	Coords_F right( m_walkingSpeedUnitsPerFrame, 0.0f );
	Coords_F noMovement( 0.0f, 0.0f );

	bool hasMoved = false;
	if( m_inputManager->IsButtonPressed( DPAD_UP ) ) {	
		myActor->MoveBy( up );
		hasMoved = true;
		m_coreUtils->Random();
	} else if( m_inputManager->IsButtonPressed( DPAD_DOWN ) ) {
		myActor->MoveBy( down );
		hasMoved = true;
	}
	
	if( m_inputManager->IsButtonPressed( DPAD_LEFT ) ) {
		float newX = myActor->GetPosition().x + left.x;
		if( newX >= -20.0f ) {
			myActor->MoveBy( left );
		} else {
			myActor->MoveBy( noMovement );
		}
		hasMoved = true;
	} else if( m_inputManager->IsButtonPressed( DPAD_RIGHT ) ) {
		float newX = myActor->GetPosition().x + right.x;
		
		if( m_scroller->CanScroll() ) {
			if( newX > SCROLLING_THRESHOLD ) {
				// I'm far enough to the right to scroll
				m_scroller->ScrollBy( right.x );
				myActor->MoveBy( noMovement );
				myActor->SetIsFacingRight( true );
			} else {
				// I'm far enough to the left to just walk to the right
				myActor->MoveBy( right );
			}
		} else {
			if( newX <= 211.0f ) {
				// I'm far enough to the left to just walk to the right
				myActor->MoveBy( right );
			} else {
				// I'm far enough to the right to not be able to walk (or scroll),
				// so just animate walking in place
				myActor->MoveBy( noMovement );
				myActor->SetIsFacingRight( true );
			}
		}
		
		hasMoved = true;
	}
	
	if( !hasMoved ) {
		myActor->Stand();
	}
}

void PlayerDriver::Decision() {
	
	m_soundsRemainingForCurrentFrame = 4;
	m_frameCounter++;
	m_character->SetLastVictim( NULL );	
	
	this->HandleSingleHit();
	this->HandleComboHit();
	this->HandleHeal();
	this->HandleMystical();
	this->HandleMovement();
}

bool PlayerDriver::TryAllocateSound() {
	if( 0 == m_soundsRemainingForCurrentFrame ) {
		return false;
	}
	
	m_soundsRemainingForCurrentFrame--;
	return true;
}