#include "SimpleEnemyDriver.h"

SimpleEnemyDriver::SimpleEnemyDriver(
	Console* console,
	CoreUtils* coreUtils,
	Character* character,
	Character* enemy,
	Impact* impact,
	bool alwaysKnockDown
	) : Driver( coreUtils, console, character, impact ) {
	
	m_enemy = enemy;
	m_alwaysKnockDown = alwaysKnockDown;
	
	m_walkingSpeedUnitsPerFrame = character->GetProperties()->GetWalkingSpeed();
	m_walkingStandingCooldownMax = 30;
	m_hittingRadius = 36.0f;
	
	m_chanceOutOfThousandToKeepWalking = character->GetProperties()->GetChanceToKeepWalking();
	m_chanceOutOfThousandToKeepNonWalking = character->GetProperties()->GetChanceToKeepNonWalking();
	m_chanceOutOfThousandToStrikeWhenClose = character->GetProperties()->GetChanceToAttack();
	
	m_wasWalkingLastFrame = false;
	m_singleHitOccurredEvent.Reset();
}

SimpleEnemyDriver::~SimpleEnemyDriver() {
}

void SimpleEnemyDriver::FarFromEnemy() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	UtilitiesCompositeActor* enemyActor = m_enemy->GetActor();
	
	// I'm not near the enemy, and I cannot move towards it
	// so there's nothing more I can do right now
	if( !myActor->CanMoveOrStand() ) {
		return;
	}
	
	bool doTheSameThingAsLastFrameRoll = m_coreUtils->Random( 1000 ) < m_chanceOutOfThousandToKeepWalking;
	if( !m_wasWalkingLastFrame ) {
		doTheSameThingAsLastFrameRoll = m_coreUtils->Random( 1000 ) < m_chanceOutOfThousandToKeepNonWalking;
	}
	
	// High chance of doing the same thing I was doing last frame
	// (walking or standing)
	//
	// or keep doing what I was doing if cooldown isn't up yet
	if( m_walkingStandingCooldown > 0 || doTheSameThingAsLastFrameRoll ) {
		
		if( m_walkingStandingCooldown > 0 ) {
			m_walkingStandingCooldown--;
		}
		
		// I wish I had C#'s lambdas...
		if( m_wasWalkingLastFrame ) {
			myActor->MoveBy( 
				myActor->DeltaTo( enemyActor, m_walkingSpeedUnitsPerFrame )
				);
			m_wasWalkingLastFrame = true;
		} else {
			myActor->Stand();
			m_wasWalkingLastFrame = false;
		}
	} else {
	
		// small chance of doing the opposite, resetting cooldown
		if( m_wasWalkingLastFrame ) {
			myActor->Stand();
			m_wasWalkingLastFrame = false;
			m_walkingStandingCooldown = m_walkingStandingCooldownMax;
		} else {
			myActor->MoveBy( 
				myActor->DeltaTo( enemyActor, m_walkingSpeedUnitsPerFrame )
				);
			m_wasWalkingLastFrame = true;
			m_walkingStandingCooldown = m_walkingStandingCooldownMax;
		}
	
	}
}

void SimpleEnemyDriver::WithinStrikingDistanceOfEnemy() {
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	UtilitiesCompositeActor* enemyActor = m_enemy->GetActor();
	
	bool canIHit = myActor->CanPerformSingleHit();
	bool canEnemyBeHit = enemyActor->CanFallDownOrRecoverFromHit();
	bool diceRoll = m_coreUtils->Random( 1000 ) < m_chanceOutOfThousandToStrikeWhenClose;
	if( diceRoll && canIHit && canEnemyBeHit ) {
		
		m_character->GetSoundboard()->StartSingleHit();
		myActor->PerformSingleHit( 
			&m_singleHitOccurredEvent, 0.5f 
			);
	} else {
		// If I'm near him but won't hit, I'll just stand if I can
		if( myActor->CanMoveOrStand() ) {
			myActor->Stand();
		}
	}
	
	return;
}

void SimpleEnemyDriver::Decision() {
	
	m_enemy->SetLastAggressor( NULL );
	UtilitiesCompositeActor* myActor = m_character->GetActor();
	UtilitiesCompositeActor* enemyActor = m_enemy->GetActor();

	// catch occurrence of a single hit event
	if( m_singleHitOccurredEvent.IsSet() ) {
		m_singleHitOccurredEvent.Reset();
		m_console->Debug("Single hit event from enemy!");
		if( myActor->IsFacingWithinDistanceOf( enemyActor, m_hittingRadius ) &&
			enemyActor->CanFallDownOrRecoverFromHit()		
			) {
			
			m_character->GetSoundboard()->SingleHitLanded();
			
			m_enemy->GetProperties()->DecreaseLifeBy( 
				m_character->GetProperties()->GetSingleHitDamage()
				);
			m_enemy->SetLastAggressor( m_character );
			bool hasDied = m_enemy->GetProperties()->GetLife() <= 0;
			
			if( hasDied || m_alwaysKnockDown ) {
				enemyActor->SetDisplacementTendency( myActor->IsToTheLeftOf( enemyActor ) );
				if( hasDied ) {
					enemyActor->SetHasDied();
				}
				m_enemy->GetSoundboard()->WasThrownDown();
				enemyActor->FallDown();
			} else {
				enemyActor->SetDisplacementTendency( myActor->IsToTheLeftOf( enemyActor ) );
				if( hasDied ) {
					enemyActor->SetHasDied();
				}
				enemyActor->RecoverFromHit();
			}
			
			m_impact->ShowForPosition( myActor->GetPosition(), myActor->IsFacingRight() );
		}
	}
	
	bool canEnemyBeHit = enemyActor->CanFallDownOrRecoverFromHit();
	if( !canEnemyBeHit ) {
		myActor->Stand();
		return;
	}
	
	// if I'm far from my enemy, get closer
	if( !myActor->IsWithinDistanceOfVerticallyAdjusted( enemyActor, m_hittingRadius ) ) {
		this->FarFromEnemy();
		return;
	}
		
	// if I'm near my enemy and facing him, then possibly strike
	if( myActor->IsFacingWithinDistanceOf( enemyActor, m_hittingRadius ) ) {	
		this->WithinStrikingDistanceOfEnemy();
		return;
	}
	
	// I'm near the enemy, but not facing him
	// turning around could be considered moving, so this check makes sense,
	// so I don't turn while hitting
	if( myActor->CanMoveOrStand() ) {
		myActor->SetIsFacingRight( !myActor->IsFacingRight() );
	}
}
