#include "Level.h"

Level::Level(
	Console* console,
	BackgroundFactory* backgroundFactory,
	SpriteManager* spriteManager,
	InputManager* inputManager,
	FighterFactory* fighterFactory,
	SpawnRegistryFactory* spawnRegistryFactory,
	PlayerViewFactory* playerViewFactory,
	EnemyViewFactory* enemyViewFactory,
	BackgroundTypes backgroundType,
	SpawnRegistryTypes spawnRegistryType,
	FrameRunnerResult previousRunnerResult,
	Impact* impact,
	Go* go,
	SoundboardFactory* soundBoardFactory
	) {
	
	m_console = console;
	m_inputManager = inputManager;
	m_spriteManager = spriteManager;
	m_fighterFactory = fighterFactory;
	m_impact = impact;
	m_go = go;
	m_soundboard = soundBoardFactory->Get();
	
	m_enemies = new List<Fighter>();
	m_fighterListEmptyFuncBool = new FighterListEmptyFuncBool( m_enemies );
	
	m_spawnRegistry = spawnRegistryFactory->Create( spawnRegistryType );
	
	m_background = backgroundFactory->Create( backgroundType, m_fighterListEmptyFuncBool );
	
	m_playerFighter = m_fighterFactory->CreatePlayer( 
		m_background,
		previousRunnerResult.first,
		previousRunnerResult.second
		);

	m_playerView = playerViewFactory->Create( m_playerFighter );
	m_enemyView = enemyViewFactory->Create( m_playerFighter );
	
	m_isPaused = false;
	m_isDone = false;
	m_outroFrames = 90;
	
	console->Debug( "Level: at %p", this );
}

Level::~Level() {

	Node<Fighter>* node = m_enemies->GetHead();
	while( NULL != node ) {
		delete node->GetData();
		node = node->GetNext();
	}
	delete m_enemies;
	
	delete m_playerFighter;
	delete m_background;
	delete m_spawnRegistry;
	
	delete m_fighterListEmptyFuncBool;
	
	delete m_playerView;
	delete m_enemyView;
	
	m_console->Debug( "~Level: at %p", this );
}

void Level::RemoveDeadEnemies() {
	Node<Fighter>* node = m_enemies->GetHead();
	while( NULL != node ) {
		Node<Fighter>* next = node->GetNext();
		
		Fighter* removalCandidate = node->GetData();
		if( removalCandidate->CanBeRemoved() ) {
			m_enemies->DeleteWhere( removalCandidate );
			delete removalCandidate;
		}
		
		node = next;
	}
}

void Level::Pre() {

	this->RemoveDeadEnemies();

	if( m_isDone && m_outroFrames == 0 ) {
		return;
	}
	
	if( m_isDone ) {
		m_outroFrames--;
		if( m_outroFrames == 0 ) {
			m_background->ResetMosaicEffect();
		}
		return;
	}
	
	m_inputManager->Scan();

	// pause/unpause?
	if( m_inputManager->IsButtonTapped( BUTTON_START ) ) {
		m_isPaused = !m_isPaused;
		m_soundboard->PauseUnpause();
		return;
	}
	
	if( m_isPaused ) {
		return;
	}
	
	// am I out of enemies?
	if( NULL == m_enemies->GetHead() ) {
		u16 currentScrollAmount = m_background->GetCurrentScrollAmount();
		
		u8 availableSpawnCount = m_spawnRegistry->GetAvailableSpawnCount( currentScrollAmount );
		if( availableSpawnCount > 0 ) {
			Spawn** availableSpawns = m_spawnRegistry->GetAvailableSpawns( currentScrollAmount );
			for( u8 i = 0; i < availableSpawnCount; i++ ) {
				Fighter* fighter = m_fighterFactory->CreateFromSpawn( 
					availableSpawns[ i ],
					m_background 
					);
				m_enemies->Add( fighter );
				m_console->Debug( "Spawning!" );
			}
			m_go->Hide();
		} else {
			m_go->Show();
		}
	}
	
	// am I still out of enemies and at the end of the background?
	if( NULL == m_enemies->GetHead() && m_background->IsAtEnd() ) {
		m_go->Hide();
		m_isDone = true;
		m_result.first = m_playerFighter->GetCharacter()->GetProperties()->GetLife();
		m_result.second = m_playerFighter->GetCharacter()->GetProperties()->GetEnergy();
	}
	
	m_playerFighter->Decision();
	
	Node<Fighter>* node = m_enemies->GetHead();
	while( NULL != node ) {
		node->GetData()->Decision();
		node = node->GetNext();
	}
	
	m_playerView->Update();
	m_enemyView->Update();
	
		// is player dead?
	if( m_playerFighter->CanBeRemoved() ) {
		m_isDone = true;
		m_result.first = 0;
		m_result.second = 0;
		return;
	}
	
	// throttle sprite reprioritization, as it can sometimes cause flickering
	if( 0 == ( this->GetGlobalFrameCounter() & 7 ) ) {
		m_spriteManager->ReprioritizeSprites();
	}
}

void Level::Frame() {

	// we always render the Go arrow, so that it can be properly hidden
	// on the first frame of the level completion countdown
	if( !m_isPaused ) {
		m_go->Render();
	}

	if( m_isDone && m_outroFrames == 0 ) {
		return;
	}
	
	if( m_isDone && m_outroFrames % 10 == 0 ) {
		m_background->IncreaseMosaicEffect();
		m_background->Render();
		return;
	}
	
	m_background->Render();
	
	if( m_isDone ) {
		return;
	}
	
	if( m_isPaused ) {
		return;
	}

	m_playerFighter->Render();
	
	Node<Fighter>* node = m_enemies->GetHead();
	while( NULL != node ) {
		node->GetData()->Render();
		node = node->GetNext();
	}
	
	m_playerView->Render();
	m_enemyView->Render();
	m_impact->Render();
}

void Level::Post() {
}

bool Level::IsDone() {
	return m_isDone && m_outroFrames == 0;
}
FrameRunnerResult Level::GetResult() {
	return m_result;
}