#include "TestsRunner.h"

TestsRunner::TestsRunner() {
	m_coreUtils = new CoreUtils();
	m_console = new Console( m_coreUtils, false );
	m_audioPlayer = new AudioPlayer( m_console, m_coreUtils );
	m_soundBoardFactory = new SoundboardFactory( m_coreUtils, m_console, m_audioPlayer );
}

TestsRunner::~TestsRunner() {
	delete m_soundBoardFactory;
	delete m_audioPlayer;
	delete m_console;
	delete m_coreUtils;
}

void TestsRunner::T_Random() {
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
		
	// test fairness by expecting at least K% within each decile
	
	u16 count = 10000;
	u16 expectedPercentageInEachDecile = 7;
	u16 deciles[ 10 ];
	
	for( int j = 0; j < 10; j++ ) {
		deciles[ j ] = 0;
	}
	
	for( int j = 0; j < count; j++ ) {
		u16 r = coreUtils->Random( 100 );
		u16 index = r / 10;
		deciles[ index ]++;
	}
	
	console->Debug( "\nDECILE DISTRIBUTION:" );
	for( int j = 0; j < 10; j++ ) {
		float decilePercentage = (float)deciles[ j ] / (float)count;
		decilePercentage = decilePercentage * 100.0f;
		console->Debug( "[%d to %d]: %d%%", j*10, j*10+9, (u16)decilePercentage );
		if( expectedPercentageInEachDecile > (u16)decilePercentage ) {
			console->Error( "PERCENTAGE TOO LOW!" );
			while( true ){}
		}
	}
	
	console->Debug( "\nDONE TESTING RANDOM NUMBERS FAIRNESS" );
	coreUtils->Sleep( 8000 );
	
	for( int j = 0; j < 100000; j++ ) {
		if( coreUtils->Random( 200 ) >= 200 ) {
			console->Error( "Random number out of range!" );
		}
	}
	
	console->Debug( "\nDONE TESTING RANDOM NUMBERS RANGE" );
	coreUtils->Sleep( 5000 );
	
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SoundEffectsPlayedMoreTimesThanItsMultiplexingLimit() {
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	AudioPlayer* m_audioPlayer = new AudioPlayer( console, coreUtils );
	
	SoundEffect* sound;
	
	sound = new SoundEffect( SFX_CHARGE1, 1, console );
	
	for(int j=0;j<5;j++ ) {
		sound->Play();
		//sound->Stop();
	}
	
	console->Debug( "\nDONE TESTING SOUND MULTIPLEXING PLAYING PAST LIMIT" );
	coreUtils->Sleep( 5000 );
	
	delete sound;
	delete m_audioPlayer;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SoundEffectsConcurrently() {
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	AudioPlayer* m_audioPlayer = new AudioPlayer( console, coreUtils );
	
	SoundEffect* sound;
	
	int multiplex = 8;
	int count = 200;
	for(int i=0;i<count;i++ ) {
		sound = new SoundEffect( SFX_CHARGE1, multiplex, console );
		
		for(int j=0;j<multiplex;j++ ) {
			sound->Play();
			sound->Stop();
		}
		
		delete sound;
	}
	sound = new SoundEffect( SFX_CHARGE1, multiplex, console );
	for(int i=0;i<count;i++ ) {
		for(int j=0;j<multiplex;j++ ) {
			sound->Play();
			sound->Stop();
		}
	}
	delete sound;
	console->Debug( "\nDONE TESTING SOUND MULTIPLEXING FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	// multiplexing test
	sound = new SoundEffect( SFX_CHARGE1, multiplex, console );
	
	for( int j = 0; j < 3; j++ ) {
		for(int i=0;i<multiplex+2;i++ ) {
			sound->Play();
			coreUtils->Sleep( 50 );
		}
		coreUtils->Sleep( 1000 );
	}
	delete sound;
	
	console->Debug( "\nDONE TESTING SOUND MULTIPLEXING OUTPUT" );
	coreUtils->Sleep( 5000 );
	
	sound = new SoundEffect( SFX_CHARGE1, 5, console );
	for( int i = 0; i < 3; i++ ) {
		
		for( int j = 0; j < 3; j++ ) {
			sound->Play();
			coreUtils->Sleep( 100 );
		}
		sound->Stop();
		
		coreUtils->Sleep( 1000 );
	}
	delete sound;
	
	console->Debug( "\nDONE TESTING SOUND MULTIPLEXING STOPPING WITH UNUSED SLOTS" );
	coreUtils->Sleep( 5000 );
	
	delete m_audioPlayer;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SongStopPlayConstructDestroy() {

	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	
	Song* m_song;
	AudioPlayer* m_audioPlayer = new AudioPlayer( console, coreUtils );
	
	for( int i=0;i< 1000;i++) {
		m_song = m_audioPlayer->LoadSong( MOD_FUNKYFIGHT );
		m_song->Play();
		m_song->Stop();
		delete m_song;
	}
	
	console->Debug( "\nDONE TESTING SONG FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	int limit = 2;
	u16 delay = 1000;
	for( int i=0;i<limit;i++) {
		m_song = m_audioPlayer->LoadSong( MOD_FUNKYFIGHT );
		m_song->Play();
		coreUtils->Sleep( delay );
		m_song->Stop();
		coreUtils->Sleep( delay );
		m_song->Play();
		coreUtils->Sleep( delay );
		m_song->Stop();
		coreUtils->Sleep( delay );
		delete m_song;
	}
	
	console->Debug( "\nDONE TESTING SONG PLAY/STOP" );
	coreUtils->Sleep( 5000 );
	
	delete m_audioPlayer;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SpriteAllocation_NoLeaks() {
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	MainScreenSpriteManager* spriteManager = new MainScreenSpriteManager( console );
	
	int spriteCount = 16;
	
	int numIterations = 100;
	OAMSprite* sprites[16];
	for( int i=0; i < numIterations; i++ ) {
	
		for( int j=0; j < spriteCount; j++ ) {
			sprites[j] = spriteManager->AllocateSprite( 
				(u8*) 1337, 
				256, 
				(u8*) 1337
				);
		}
		
		for( int j=0; j < spriteCount; j++ ) {
			delete sprites[j];
		}
	}
	
	console->Debug( "\nDONE TESTING SPRITE ALLOCATION FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	delete spriteManager;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SubSpriteAllocation_NoLeaks() {
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	SubScreenSpriteManager* spriteManager = new SubScreenSpriteManager( console );
	
	int spriteCount = 32;
	
	int numIterations = 50;
	OAMSprite* sprites[32];
	for( int i=0; i < numIterations; i++ ) {
	
		for( int j=0; j < spriteCount; j++ ) {
			sprites[j] = spriteManager->AllocateSprite();
		}
		
		for( int j=0; j < spriteCount; j++ ) {
			delete sprites[j];
		}
	}
	
	console->Debug( "\nDONE TESTING SUB SCREEN SPRITE ALLOCATION FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	delete spriteManager;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SpritePoolReprioritization() {
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	
	int spriteCount = 16;
	
	int numIterations = 50;
	OAMSprite* sprites[spriteCount];
	
	for( int j=0; j < numIterations; j++ ) {
	
		TestsSpritePool* pool = new TestsSpritePool( console, spriteCount );
	
		for( int i=0; i < spriteCount; i++ ) {
			sprites[i] = new OAMSprite( 
				i,
				(u8*) 1337, 
				256, 
				pool,
				console,
				&oamMain
				);
				
			pool->AddSprite( sprites[i] );
			
			Coords_F spriteCoords( 
				coreUtils->Random( 190 ), 
				coreUtils->Random( 190 ) 
				);
			
			sprites[i]->MoveTo( spriteCoords );
		}
		
		// create a few empty slots
		delete sprites[ 4 ];
		delete sprites[ 8 ];
		
		pool->Reprioritize();

		Coords_F moveSpriteCoords( 
			coreUtils->Random( 190 ), 
			coreUtils->Random( 190 ) 
			);
		
		if( 0 == j % 2 ) {
			sprites[1]->MoveTo( moveSpriteCoords );
			pool->Reprioritize();
		}
		
		OAMSprite** poolArray = pool->GetSpritesArray();
		
		for( int i=0; i < spriteCount; i++ ) {
			
			if( NULL == poolArray[ i ] ) {
				break;
			}
			
			Coords_F coords( 
				coreUtils->Random( 190 ), 
				coreUtils->Random( 190 ) 
				);
			
			sprites[i]->MoveTo( coords );
		}
		
		pool->Reprioritize();
		
		u8 contiguousNonNullCount = 0;
		for( int i=0; i < spriteCount; i++ ) {
			
			if( NULL == poolArray[ i ] ) {
				break;
			}
			contiguousNonNullCount++;
		}
		
		if( contiguousNonNullCount != spriteCount - 2 ) {
			console->Error( "Pool not compacted!" );
		}
		
		for( int i=0; i < contiguousNonNullCount-1; i++ ) {
			
			Coords_F currentPosition = poolArray[ i ]->GetPosition();
			Coords_F nextPosition = poolArray[ i + 1 ]->GetPosition();
			
			if( currentPosition.y < nextPosition.y ) {
				console->Error( "Pool not properly prioritized!" );
			}
		}
		
		delete pool;
	}
	
	console->Debug( "\nDONE TESTING SPRITE POOL REPRIORITIZATION ON SPRITE MOVEMENT" );
	coreUtils->Sleep( 5000 );
	
	delete console;
	delete coreUtils;
}

void TestsRunner::T_SpritePoolReprioritizedOnFreeSlot() {
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	
	int spriteCount = 16;

	OAMSprite* sprites[spriteCount];
	
	TestsSpritePool* pool = new TestsSpritePool( console, spriteCount );

	for( int i=0; i < spriteCount; i++ ) {
		sprites[i] = new OAMSprite( 
			i,
			(u8*) 1337, 
			256, 
			pool,
			console,
			&oamMain
			);
			
		pool->AddSprite( sprites[i] );
	}
	
	delete sprites[ 4 ];
	
	OAMSprite** poolArray = pool->GetSpritesArray();
	
	u8 contiguousNonNullCount = 0;
	for( int i=0; i < spriteCount; i++ ) {
		
		if( NULL == poolArray[ i ] ) {
			break;
		}
		contiguousNonNullCount++;
	}
	
	if( contiguousNonNullCount != spriteCount - 1 ) {
		console->Error( "Freeing a slot did not re-compact the pool" );
	}
	
	console->Debug( "\nDONE TESTING SPRITE POOL RE-COMPACTING ON FREEING SLOT" );
	coreUtils->Sleep( 5000 );
	
	delete pool;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_Actors() {
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	
	SpriteManager* spriteManager = new SpriteManager( console );
	
	AnimatorFactory* animatorFactory = new AnimatorFactory( console, spriteManager );
	
	FallingDownActorFactory* fallingDownActorFactory = new FallingDownActorFactory( coreUtils, console );
	SingleActionActorFactory* singleActionActorFactory = new SingleActionActorFactory( coreUtils, console );
	HitRecoveringActorFactory* hitRecoveringActorFactory = new HitRecoveringActorFactory( console, singleActionActorFactory );
	ComboActorFactory* comboActorFactory = new ComboActorFactory( console, singleActionActorFactory );
	ChargingActorFactory* chargingActorFactory = new ChargingActorFactory( console );
	
	CompositeActorFactory* compositeActorFactory = new CompositeActorFactory( 
		coreUtils,
		console, 
		animatorFactory,
		fallingDownActorFactory,
		hitRecoveringActorFactory,
		singleActionActorFactory,
		comboActorFactory,
		chargingActorFactory
		);
	
	CompositeActor* actor;
	Event continuationEvent;
	continuationEvent.Set();
	
	int actorCount = 300;
	for( int i = 0; i < actorCount; i++ ) {
		actor = compositeActorFactory->Create( ACTOR_PLAYER );
		int waitCount = 50;
		
		
		for( int j = 0; j < waitCount; j++ ) {
			actor->FallDown();
			actor->Render();
		}
		
		for( int j = 0; j < waitCount; j++ ) {
			actor->PerformCombo( &continuationEvent, NULL, 120.0f );
			actor->Render();
		}
		
		for( int j = 0; j < waitCount; j++ ) {
			actor->PerformSingleHit( NULL, 0.0f );
			actor->Render();
		}
		
		for( int j = 0; j < waitCount; j++ ) {
			actor->RecoverFromHit();
			actor->Render();
		}
		delete actor;
	}
	
	console->Debug( "\nDONE TESTING ACTORS FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	delete compositeActorFactory;
	delete animatorFactory;
	delete spriteManager;
	delete console;
	delete coreUtils;
}

void TestsRunner::T_Levels() {
	
	CoreUtils* coreUtils = new CoreUtils();
	Console* console = new TestsConsole( coreUtils );
	
	LevelFactory* levelFactory = new LevelFactory(
		console,
		coreUtils,
		m_soundBoardFactory
		);
	FrameRunnerResult previousResult( 10, 10 );
	
	int count = 300;
	for( int i = 0; i < count; i++ ) {
		Level* level = levelFactory->Create( CIRCUS_PARK_BUILDINGS, SR_CIRCUS_PARK_BUILDINGS, previousResult );
		delete level;
	}
	
	delete levelFactory;
	
	console->Debug( "\nDONE TESTING LEVELS FOR MEMORY LEAKS" );
	coreUtils->Sleep( 5000 );
	
	delete console;
	delete coreUtils;
}

void TestsRunner::Start() {
	SystemContext* context = new SystemContext();
	context->Initialize();
	
	CoreUtils* coreUtils = new CoreUtils();
	TestsConsole* console = new TestsConsole( coreUtils );
	console->Debug( "TESTS ARE ABOUT TO BEGIN" );
	console->Debug( "\nLOWER THE VOLUME !" );
	coreUtils->Sleep( 5000 );
	
	T_SoundEffectsPlayedMoreTimesThanItsMultiplexingLimit();
	T_Levels();
	T_Actors();
	T_Random();
	T_SoundEffectsConcurrently();
	T_SongStopPlayConstructDestroy();
	T_SpriteAllocation_NoLeaks();
	T_SubSpriteAllocation_NoLeaks();
	T_SpritePoolReprioritization();
	T_SpritePoolReprioritizedOnFreeSlot();

	console->Debug( "\nTEST SUITE SUCCESSFUL !" );
	console->Debug( "\nCPU HALTED" );
	
	while( true ) {}
}